"""
This module provides functionalities to represent and create mixtures of spherical particles
with various size distributions. It includes classes and functions to handle different
distributions like single-component, Flory-Schulz, Gaussian, and uniform distributions.

Classes:
    Mixture: Base class representing an N-component mixture of spherical particles.
    FlorySchulzMixture: Represents a Schulz-Flory distributed mixture.
    GaussianMixture: Represents a Gaussian distributed mixture.
    UniformMixture: Represents a uniformly distributed mixture.
    SingleComponent: Represents a pseudo one-component system.
"""

from .mixture import Mixture, FlorySchulzMixture, GaussianMixture, UniformMixture, SingleComponent

__all__ = ["Mixture", "FlorySchulzMixture", "GaussianMixture", "UniformMixture", "SingleComponent"]



"""
Module `liquidstructure` provides classes for representing the liquid structure properties of
interacting multicomponent systems of spherical particles.

Classes:
    MixtureLike: A protocol defining the expected interface for mixture-like objects.
    LiquidStructure: Abstract base class representing liquid structure properties.
    PercusYevick: Class for hard-sphere potential in the Percus-Yevick approximation.
    VerletWeis: Class for hard-sphere potential in the Percus-Yevick approximation
        with Verlet-Weiss correction.

References:
    - P. Salgi, R. Rajagopalan, "Polydispersity in colloids: implications to static structure and scattering",
      Adv. Colloid Interface Sci. 43, 169-288 (1993), https://doi.org/10.1016/0001-8686(93)80017-6
    - A. Vrij, "Mixtures of hard spheres in the Percus–Yevick approximation. Light scattering at finite angles",
      J. Chem. Phys. 71, 3267-3270 (1979), https://doi.org/10.1063/1.438756
    - J. Diaz Maier, K. Gaus, J. Wagner, "Measurable structure factors of dense dispersions containing polydisperse,
      optically inhomogeneous particles", arXiv:2404.03470 [cond-mat.soft], https://doi.org/10.48550/arXiv.2404.03470
"""

from .liquidstructure import LiquidStructure, PercusYevick, VerletWeis

__all__ = ["LiquidStructure", "PercusYevick", "VerletWeis"]



# type: ignore
import pytest
import numpy as np
from numpy.testing import assert_array_almost_equal
from mixscatter.scatteringmodel import (
    EmptyProfile,
    ConstantProfile,
    LinearProfile,
    Particle,
    ParticleBuilder,
    ScatteringModel,
    SimpleSphere,
    SimpleCoreShell,
    SimpleGradient,
)


@pytest.fixture
def wavevector():
    return np.array([0.1, 1.0, 10.0])


@pytest.fixture
def distance():
    return np.array([0.1, 0.5, 1.0, 10.0])


@pytest.fixture
def mock_mixture(mocker):
    mock = mocker.Mock()
    mock.number_fraction = np.array([0.5, 0.5])
    mock.radius = np.array([1.0, 2.0])
    return mock


def test_empty_profile(wavevector, distance):
    profile = EmptyProfile(0, 1)
    amplitude = profile.calculate_amplitude(wavevector)
    forward_amplitude = profile.calculate_forward_amplitude()
    layer_profile = profile.get_profile(distance)

    assert_array_almost_equal(amplitude, np.zeros_like(wavevector))
    assert forward_amplitude == 0.0
    assert_array_almost_equal(layer_profile, np.zeros_like(distance))


def test_empty_profile_invalid_radius():
    with pytest.raises(RuntimeError, match="'radius_inner' must be smaller than 'radius_outer'."):
        EmptyProfile(1, 0)


def test_constant_profile(wavevector, distance):
    profile = ConstantProfile(0, 1, 1.0)
    amplitude = profile.calculate_amplitude(wavevector)
    forward_amplitude = profile.calculate_forward_amplitude()
    layer_profile = profile.get_profile(distance)

    expected_forward_amplitude = 4.0 / 3.0 * np.pi * (1.0**3 - 0.0**3)

    assert_array_almost_equal(amplitude, np.array([4.185, 3.785, 0.099]), decimal=3)
    assert np.isclose(forward_amplitude, expected_forward_amplitude, atol=1e-5)
    assert_array_almost_equal(layer_profile, np.array([1.0, 1.0, 0.0, 0.0]), decimal=3)


def test_constant_profile_invalid_radius():
    with pytest.raises(RuntimeError, match="'radius_inner' must be smaller than 'radius_outer'."):
        ConstantProfile(1, 0, 1.0)


def test_linear_profile(wavevector, distance):
    profile = LinearProfile(0, 1, 1.0, 2.0)
    amplitude = profile.calculate_amplitude(wavevector)
    forward_amplitude = profile.calculate_forward_amplitude()
    layer_profile = profile.get_profile(distance)

    expected_forward_amplitude = 4.0 / 3.0 * np.pi * (1.0**3 - 0.0**3) + np.pi * (1.0**4 - 0.0**4)

    assert_array_almost_equal(amplitude, np.array([7.323, 6.590, 0.186]), decimal=3)
    assert np.isclose(forward_amplitude, expected_forward_amplitude, atol=1e-5)
    assert_array_almost_equal(layer_profile, np.array([1.1, 1.5, 0.0, 0.0]), decimal=3)


def test_linear_profile_invalid_radius():
    with pytest.raises(RuntimeError, match="'radius_inner' must be smaller than 'radius_outer'."):
        LinearProfile(1, 0, 1.0, 2.0)


def test_particle(wavevector):
    profile1 = ConstantProfile(0, 1, 1.0)
    profile2 = LinearProfile(1, 2, 1.0, 2.0)
    particle = Particle([profile1, profile2])

    amplitude = particle.calculate_amplitude(wavevector)
    forward_amplitude = particle.calculate_forward_amplitude()
    form_factor = particle.calculate_form_factor(wavevector)
    distance = np.array([0, 1, 1.5, 2, 10])
    layer_profile = particle.get_profile(distance)

    assert_array_almost_equal(amplitude, np.array([51.089, 32.012, -0.149]), decimal=3)
    assert np.isclose(forward_amplitude, 51.313, atol=1e-3)
    assert_array_almost_equal(form_factor, np.array([9.913e-01, 3.892e-01, 8.462e-06]), decimal=3)
    assert_array_almost_equal(layer_profile, np.array([1.0, 1.0, 1.5, 0.0, 0.0]), decimal=3)


def test_particle_layer_connection():
    profile1 = EmptyProfile(0, 1)
    profile2 = ConstantProfile(1, 2, 1.0)
    profile3 = LinearProfile(2, 3, 5.0, 2.0)
    profile4 = EmptyProfile(3, 4)
    particle = Particle([profile1, profile2, profile3, profile4])

    assert_array_almost_equal(particle.get_profile([1.0, 2.0, 3.0]), np.array([1.0, 5.0, 0.0]))


def test_particle_builder():
    builder = ParticleBuilder()
    builder.add_layer(ConstantProfile(0, 1, 1.0))
    builder.add_layer(ConstantProfile(1, 2, 2.0))
    particle = builder.get_particle()

    assert len(particle.layers) == 2


def test_particle_builder_invalid_layer_connection():
    builder = ParticleBuilder()
    layer1 = ConstantProfile(0, 1, 1.0)
    layer2 = ConstantProfile(0.5, 2, 2.0)

    builder.add_layer(layer1)

    with pytest.raises(RuntimeError) as excinfo:
        builder.add_layer(layer2)

    expected_message = f"Added layer with inner radius {layer2.radius_inner} does not connect to previous layer's outer radius {layer1.radius_outer}."
    assert str(excinfo.value) == expected_message


def test_simple_sphere(wavevector, mock_mixture):
    model = SimpleSphere(wavevector, mock_mixture, 1.0)
    assert model.amplitude.shape == (2, len(wavevector))
    assert model.forward_amplitude.shape == (2,)


def test_simple_core_shell(wavevector, mock_mixture):
    model = SimpleCoreShell(wavevector, mock_mixture, 0.5, 1.0, 2.0)
    assert model.amplitude.shape == (2, len(wavevector))
    assert model.forward_amplitude.shape == (2,)


def test_simple_gradient(wavevector, mock_mixture):
    model = SimpleGradient(wavevector, mock_mixture, 1.0, 2.0)
    assert model.amplitude.shape == (2, len(wavevector))
    assert model.forward_amplitude.shape == (2,)


def test_scattering_model_properties(wavevector, mock_mixture):
    particles = [Particle([ConstantProfile(0, 1, 1.0)]), Particle([LinearProfile(0, 1, 1.0, 2.0)])]
    model = ScatteringModel(wavevector, mock_mixture, particles)

    assert model.amplitude.shape == (2, len(wavevector))
    assert model.forward_amplitude.shape == (2,)
    assert model.single_form_factor.shape == (2, len(wavevector))
    assert model.average_square_amplitude.shape == (len(wavevector),)
    assert np.isscalar(model.average_square_forward_amplitude)
    assert model.average_form_factor.shape == (len(wavevector),)



"""Calculate form factors of multicomponent systems.

This module provides classes and functions for calculating form factors of multicomponent systems.
It includes definitions for various layer profiles, particles composed of these layers, and models
to calculate scattering properties for these particles.

Classes:
    LayerProfile: Defines the interface for layer profiles.
    EmptyProfile: Represents an empty layer profile.
    ConstantProfile: Represents a layer profile with constant contrast.
    LinearProfile: Represents a layer profile with linearly varying contrast.
    Particle: Represents a particle composed of multiple layers.
    ParticleBuilder: A builder class for constructing Particle instances.
    ScatteringModel: Calculates scattering properties for a list of particles.
    SimpleSphere: A convenience class for creating a scattering model of homogeneously scattering spheres.
    SimpleCoreShell: A convenience class for creating a scattering model of core-shell particles.
    SimpleGradient: A convenience class for creating a scattering model with gradient profiles.

Functions:
    _two_point_to_slope_intercept: Utility function for calculating slope and intercept between two points.

Examples:
    Create a simple scattering model for spheres:

    >>> from mixscatter.mixture import Mixture
    >>> wavevector = np.linspace(0.01, 1.0, 100)
    >>> mixture = Mixture(number_fraction=[0.5, 0.5], radius=[1.0, 2.0])
    >>> model = SimpleSphere(wavevector, mixture, contrast=1.0)
    >>> form_factor = model.average_form_factor

    Create a core-shell particle model:

    >>> model = SimpleCoreShell(wavevector, mixture, core_to_total_ratio=0.5, core_contrast=1.0, shell_contrast=0.5)
    >>> form_factor = model.average_form_factor

    Create a gradient profile particle model:

    >>> model = SimpleGradient(wavevector, mixture, center_contrast=1.0, boundary_contrast=0.5)
    >>> form_factor = model.average_form_factor
"""

from functools import cached_property
from typing import Any, Protocol

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import numpy as np
from numpy.typing import ArrayLike, NDArray


# noinspection PyPropertyDefinition
class MixtureLike(Protocol):  # pragma: no cover
    @property
    def number_fraction(self) -> NDArray[np.float64]: ...

    @property
    def radius(self) -> NDArray[np.float64]: ...


class LayerProfile(Protocol):  # pragma: no cover
    """Defines the interface for layer profiles."""

    radius_inner: float
    radius_outer: float

    def calculate_amplitude(self, wavevector: ArrayLike) -> NDArray[np.float64]:
        """Calculate the amplitude for the given wavevector."""
        ...

    def calculate_forward_amplitude(self) -> float:
        """Calculate the forward amplitude."""
        ...

    def get_profile(self, distance: ArrayLike) -> NDArray[np.float64]:
        """Get the profile for the given distance from the origin."""
        ...


class EmptyProfile(LayerProfile):
    """Represents an empty layer profile."""

    def __init__(self, radius_inner: float, radius_outer: float) -> None:
        if radius_inner > radius_outer:
            raise RuntimeError("'radius_inner' must be smaller than 'radius_outer'.")

        self.radius_inner = radius_inner
        self.radius_outer = radius_outer

    def calculate_amplitude(self, wavevector: ArrayLike) -> NDArray[np.float64]:
        """Calculate the amplitude for the given wavevector.

        Returns:
            A zero array of the same shape as the wavevector.
        """
        wavevector = np.asarray(wavevector, dtype=np.float64)
        return np.zeros_like(wavevector)

    def calculate_forward_amplitude(self) -> float:
        """Calculate the forward amplitude.

        Returns:
            Zero.
        """
        return 0.0

    def get_profile(self, distance: ArrayLike) -> NDArray[np.float64]:
        """Get the profile for the given distance from the origin.

        Returns:
            A zero array of the same shape as the distance.
        """
        distance = np.asarray(distance, dtype=np.float64)
        return np.zeros_like(distance)


class ConstantProfile(LayerProfile):
    """Represents a layer profile with constant contrast."""

    def __init__(self, radius_inner: float, radius_outer: float, contrast: float) -> None:
        if radius_inner > radius_outer:
            raise RuntimeError("'radius_inner' must be smaller than 'radius_outer'.")

        self.radius_inner = radius_inner
        self.radius_outer = radius_outer
        self.contrast = contrast

    def calculate_amplitude(self, wavevector: ArrayLike) -> NDArray[np.float64]:
        """Calculate the amplitude for the given wavevector.

        Returns:
            The calculated amplitude array.
        """
        wavevector = np.asarray(wavevector, dtype=np.float64)
        QR_outer = wavevector * self.radius_outer
        QR_inner = wavevector * self.radius_inner
        amplitude: NDArray[np.float64] = np.sin(QR_outer) - QR_outer * np.cos(QR_outer)
        amplitude -= np.sin(QR_inner) - QR_inner * np.cos(QR_inner)
        amplitude *= 4.0 * np.pi / wavevector**3 * self.contrast
        return amplitude

    def calculate_forward_amplitude(self) -> float:
        """Calculate the forward amplitude.

        Returns:
            The calculated forward amplitude.
        """
        return 4.0 / 3.0 * np.pi * (self.radius_outer**3 - self.radius_inner**3) * self.contrast

    def get_profile(self, distance: ArrayLike) -> NDArray[np.float64]:
        """Get the profile for the given distance from the origin.

        Returns:
            The profile evaluated on the distance array.
        """
        distance = np.asarray(distance, dtype=np.float64)
        distance_mask = (distance >= self.radius_inner) & (distance < self.radius_outer)
        return np.where(distance_mask, self.contrast, 0.0)


class LinearProfile(LayerProfile):
    """Represents a layer profile with linearly varying contrast."""

    def __init__(self, radius_inner: float, radius_outer: float, contrast_inner: float, contrast_outer: float) -> None:
        if radius_inner > radius_outer:
            raise RuntimeError("'radius_inner' must be smaller than 'radius_outer'.")

        self.radius_inner = radius_inner
        self.radius_outer = radius_outer
        self.contrast_inner = contrast_inner
        self.contrast_outer = contrast_outer

    def calculate_amplitude(self, wavevector: ArrayLike) -> NDArray[np.float64]:
        """Calculate the amplitude for the given wavevector.

        Returns:
            The calculated amplitude array.
        """
        wavevector = np.asarray(wavevector, dtype=np.float64)
        QR_outer = wavevector * self.radius_outer
        QR_inner = wavevector * self.radius_inner
        intercept, slope = self._two_point_to_slope_intercept(
            self.radius_inner, self.contrast_inner, self.radius_outer, self.contrast_outer
        )
        amplitude_intercept = ConstantProfile(self.radius_inner, self.radius_outer, intercept).calculate_amplitude(
            wavevector
        )
        amplitude_gradient: NDArray[np.float64] = (2.0 - QR_outer**2) * np.cos(QR_outer) + 2.0 * QR_outer * np.sin(
            QR_outer
        )
        amplitude_gradient -= (2.0 - QR_inner**2) * np.cos(QR_inner) + 2.0 * QR_inner * np.sin(QR_inner)
        amplitude_gradient *= 4.0 * np.pi / wavevector**4 * slope
        return amplitude_intercept + amplitude_gradient

    def calculate_forward_amplitude(self) -> float:
        """Calculate the forward amplitude.

        Returns:
            The calculated forward amplitude.
        """
        intercept, slope = self._two_point_to_slope_intercept(
            self.radius_inner, self.contrast_inner, self.radius_outer, self.contrast_outer
        )
        forward_amplitude_intercept = ConstantProfile(
            self.radius_inner, self.radius_outer, intercept
        ).calculate_forward_amplitude()
        forward_amplitude_gradient = np.pi * (self.radius_outer**4 - self.radius_inner**4) * slope
        return forward_amplitude_intercept + forward_amplitude_gradient

    @staticmethod
    def _two_point_to_slope_intercept(x_1: float, y_1: float, x_2: float, y_2: float) -> tuple[float, float]:
        """Calculate the slope and intercept for a linear function passing through two points.

        Returns:
            A tuple containing the intercept and slope.
        """
        intercept = (x_2 * y_1 - x_1 * y_2) / (x_2 - x_1)
        slope = (y_2 - y_1) / (x_2 - x_1)
        return intercept, slope

    def get_profile(self, distance: ArrayLike) -> NDArray[np.float64]:
        """Get the profile for the given distance from the origin.

        Returns:
            The profile evaluated on the distance array.
        """
        distance = np.asarray(distance, dtype=np.float64)
        distance_mask = (distance >= self.radius_inner) & (distance < self.radius_outer)
        intercept, slope = self._two_point_to_slope_intercept(
            self.radius_inner, self.contrast_inner, self.radius_outer, self.contrast_outer
        )
        profile = np.zeros_like(distance)
        profile[distance_mask] = intercept + slope * distance[distance_mask]
        return profile


class Particle:
    """Represents a particle composed of multiple layers."""

    def __init__(self, layers: list[LayerProfile]) -> None:
        self.layers = layers

    def calculate_amplitude(self, wavevector: ArrayLike) -> NDArray[np.float64]:
        """Calculate the amplitude for the given wavevector.

        Returns:
            The calculated amplitude array.
        """
        wavevector = np.asarray(wavevector, dtype=np.float64)
        amplitude = np.zeros_like(wavevector)
        for layer in self.layers:
            amplitude += layer.calculate_amplitude(wavevector)
        return amplitude

    def calculate_forward_amplitude(self) -> float:
        """Calculate the forward amplitude.

        Returns:
            The calculated forward amplitude.
        """
        forward_amplitude = 0.0
        for layer in self.layers:
            forward_amplitude += layer.calculate_forward_amplitude()
        return forward_amplitude

    def calculate_form_factor(self, wavevector: ArrayLike) -> NDArray[np.float64]:
        """Calculate the form factor for the given wavevector.

        Returns:
            The calculated form factor array.
        """
        wavevector = np.asarray(wavevector, dtype=np.float64)
        amplitude = self.calculate_amplitude(wavevector)
        forward_amplitude = self.calculate_forward_amplitude()
        return (amplitude / forward_amplitude) ** 2

    def get_profile(self, distance: ArrayLike) -> NDArray[np.float64]:
        """Get the profile for the given distance from the origin.

        Returns:
            The profile evaluated on the distance array.
        """
        distance = np.asarray(distance, dtype=np.float64)
        profile = np.zeros_like(distance)
        for layer in self.layers:
            profile += layer.get_profile(distance)
        return profile


class ParticleBuilder:
    """A builder class for constructing Particle instances."""

    def __init__(self) -> None:
        self.layers: list[LayerProfile] = []

    def add_layer(self, layer: LayerProfile) -> Self:
        """Add a layer to the particle being built.

        Args:
            layer: LayerProfile instance to be added.

        Returns:
            The builder instance.

        Raises:
            RuntimeError: If the added layer does not connect to the previous layer or if layers overlap.
        """
        if self.layers:
            do_layers_connect = np.allclose(self.layers[-1].radius_outer, layer.radius_inner)
            if not do_layers_connect:
                raise RuntimeError(
                    f"Added layer with inner radius {layer.radius_inner} does not connect to previous layer's outer "
                    f"radius {self.layers[-1].radius_outer}."
                )
        self.layers.append(layer)
        return self

    def reset(self) -> None:
        """Reset the builder, clearing all layers."""
        self.layers = []

    def get_particle(self) -> Particle:
        """Get the constructed particle.

        Returns:
            The constructed Particle instance.
        """
        return Particle(self.layers)

    def pop_particle(self) -> Particle:
        """Get the constructed particle and reset the builder.

        Returns:
            The constructed Particle instance.
        """
        particle = Particle(self.layers)
        self.reset()
        return particle


class ScatteringModel:
    """Calculates scattering properties for a list of particles."""

    def __init__(self, wavevector: ArrayLike, mixture: MixtureLike, particles: list[Particle]):
        self.wavevector = np.asarray(wavevector, dtype=np.float64)
        self.mixture = mixture
        self.particles = particles

    @cached_property
    def amplitude(self) -> NDArray[np.float64]:
        """Calculate the amplitude for each particle.

        Returns:
            An array of amplitudes for each particle.
        """
        amplitude = np.empty((len(self.particles), len(self.wavevector)))
        for i, particle in enumerate(self.particles):
            amplitude[i] = particle.calculate_amplitude(self.wavevector)
        return amplitude

    @cached_property
    def forward_amplitude(self) -> NDArray[np.float64]:
        """Calculate the forward amplitude for each particle.

        Returns:
            An array of forward amplitudes for each particle.
        """
        forward_amplitude = np.empty(len(self.particles))
        for i, particle in enumerate(self.particles):
            forward_amplitude[i] = particle.calculate_forward_amplitude()
        return forward_amplitude

    @cached_property
    def single_form_factor(self) -> NDArray[np.float64]:
        """The normalized form factors of the single species.

        Returns:
            The form factors.
        """
        return self.amplitude**2 / self.forward_amplitude[:, np.newaxis] ** 2

    @cached_property
    def average_square_amplitude(self) -> Any:
        """The sum of the squared scattering amplitudes, weighted by the number fraction.

        Returns:
            The average squared scattering amplitude.
        """
        return np.sum(self.mixture.number_fraction[:, np.newaxis] * self.amplitude**2, axis=0)

    @cached_property
    def average_square_forward_amplitude(self) -> Any:
        """The sum of the squared forward scattering amplitudes, weighted by the number fraction.

        Returns:
            The average squared forward scattering amplitude.
        """
        return np.sum(self.mixture.number_fraction * self.forward_amplitude**2, axis=0)

    @cached_property
    def average_form_factor(self) -> Any:
        """The average squared scattering amplitude, normalized by the average forward scattering amplitude.

        Returns:
            The average form factor.
        """
        return self.average_square_amplitude / self.average_square_forward_amplitude


class SimpleSphere(ScatteringModel):
    """A convenience class for creating a scattering model of homogeneously scattering spheres."""

    def __init__(self, wavevector: ArrayLike, mixture: MixtureLike, contrast: float) -> None:
        particles = []
        particle_builder = ParticleBuilder()
        for radius in mixture.radius:
            particle = particle_builder.add_layer(ConstantProfile(0, radius, contrast)).pop_particle()
            particles.append(particle)
        super().__init__(wavevector, mixture, particles)


class SimpleCoreShell(ScatteringModel):
    """A convenience class for creating a scattering model of core-shell particles."""

    def __init__(
        self,
        wavevector: ArrayLike,
        mixture: MixtureLike,
        core_to_total_ratio: float,
        core_contrast: float,
        shell_contrast: float,
    ) -> None:
        particles = []
        particle_builder = ParticleBuilder()
        for total_radius in mixture.radius:
            core_radius = total_radius * core_to_total_ratio
            particle = (
                particle_builder.add_layer(ConstantProfile(0, core_radius, core_contrast))
                .add_layer(ConstantProfile(core_radius, total_radius, shell_contrast))
                .pop_particle()
            )
            particles.append(particle)
        super().__init__(wavevector, mixture, particles)


class SimpleGradient(ScatteringModel):
    """A convenience class for creating a scattering model with gradient profiles."""

    def __init__(
        self,
        wavevector: ArrayLike,
        mixture: MixtureLike,
        center_contrast: float,
        boundary_contrast: float,
    ) -> None:
        particles = []
        particle_builder = ParticleBuilder()
        for radius in mixture.radius:
            particle = particle_builder.add_layer(
                LinearProfile(0, radius, center_contrast, boundary_contrast)
            ).pop_particle()
            particles.append(particle)
        super().__init__(wavevector, mixture, particles)



# type: ignore
import pytest
import numpy as np
from numpy.testing import assert_array_almost_equal

from mixscatter.liquidstructure import LiquidStructure, PercusYevick, VerletWeis


@pytest.fixture
def mock_mixture(mocker):
    mock_mixture = mocker.Mock()
    mock_mixture.number_fraction = np.array([0.5, 0.5])
    mock_mixture.radius = np.array([1.0, 2.0])
    mock_mixture.number_of_components = 2
    mock_mixture.moment.side_effect = lambda order: np.sum(mock_mixture.number_fraction * mock_mixture.radius**order)

    return mock_mixture


def test_liquid_structure_initialization(mock_mixture):
    wavevector = [0.1, 0.2, 0.3]
    with pytest.raises(TypeError):
        _ = LiquidStructure(wavevector, mock_mixture)


def test_percus_yevick_initialization(mock_mixture):
    wavevector = [0.1, 0.2, 0.3]
    volume_fraction_total = 0.3
    py = PercusYevick(wavevector, mock_mixture, volume_fraction_total)
    assert np.allclose(py.wavevector, [0.1, 0.2, 0.3])
    assert py.mixture == mock_mixture
    assert py.volume_fraction_total == volume_fraction_total


def test_percus_yevick_partial_direct_correlation_function(mock_mixture):
    wavevector = [0.1, 0.2, 0.3]
    volume_fraction_total = 0.3
    py = PercusYevick(wavevector, mock_mixture, volume_fraction_total)
    c_ijq = py.partial_direct_correlation_function
    expected_c_ijq = np.array(
        [[[-1.316, -1.302, -1.278], [-5.417, -5.287, -5.074]], [[-5.417, -5.287, -5.074], [-19.565, -18.804, -17.587]]]
    )
    assert c_ijq.shape == (2, 2, 3)
    assert_array_almost_equal(c_ijq, expected_c_ijq, decimal=3)


def test_percus_yevick_number_weighted_partial_direct_correlation_function(mock_mixture):
    wavevector = [0.1, 0.2, 0.3]
    volume_fraction_total = 0.3
    py = PercusYevick(wavevector, mock_mixture, volume_fraction_total)
    c_weighted_ijq = py.number_weighted_partial_direct_correlation_function
    expected_c_weighted_ijq = np.array(
        [[[-0.658, -0.651, -0.639], [-2.709, -2.643, -2.537]], [[-2.709, -2.643, -2.537], [-9.783, -9.402, -8.793]]]
    )
    assert c_weighted_ijq.shape == (2, 2, 3)
    assert_array_almost_equal(c_weighted_ijq, expected_c_weighted_ijq, decimal=3)


def test_percus_yevick_partial_structure_factor(mock_mixture):
    wavevector = [0.1, 0.2, 0.3]
    volume_fraction_total = 0.3
    py = PercusYevick(wavevector, mock_mixture, volume_fraction_total)
    S_ijq = py.partial_structure_factor
    expected_S_ijq = np.array(
        [[[1.023, 1.021, 1.02], [-0.257, -0.260, -0.264]], [[-0.257, -0.260, -0.264], [0.157, 0.162, 0.170]]]
    )
    assert S_ijq.shape == (2, 2, 3)
    assert_array_almost_equal(S_ijq, expected_S_ijq, decimal=3)


def test_percus_yevick_number_weighted_partial_structure_factor(mock_mixture):
    wavevector = [0.1, 0.2, 0.3]
    volume_fraction_total = 0.3
    py = PercusYevick(wavevector, mock_mixture, volume_fraction_total)
    S_weighted_ijq = py.number_weighted_partial_structure_factor
    expected_S_weighted_ijq = np.array(
        [[[0.511, 0.511, 0.509], [-0.128, -0.130, -0.132]], [[-0.128, -0.130, -0.132], [0.079, 0.081, 0.085]]]
    )
    assert S_weighted_ijq.shape == (2, 2, 3)
    assert_array_almost_equal(S_weighted_ijq, expected_S_weighted_ijq, decimal=3)


def test_percus_yevick_average_structure_factor(mock_mixture):
    wavevector = [0.3, 1.9, 10.0]
    volume_fraction_total = 0.5
    py = PercusYevick(wavevector, mock_mixture, volume_fraction_total)
    S_avg_q = py.average_structure_factor
    assert S_avg_q.shape == (3,)
    assert_array_almost_equal(S_avg_q, np.array([0.344, 1.508, 0.994]), decimal=3)


def test_percus_yevick_compressibility_structure_factor(mock_mixture):
    wavevector = [0.3, 1.9, 10.0]
    volume_fraction_total = 0.5
    py = PercusYevick(wavevector, mock_mixture, volume_fraction_total)
    S_compressibility_q = py.compressibility_structure_factor
    assert S_compressibility_q.shape == (3,)
    assert_array_almost_equal(S_compressibility_q, np.array([0.021, 1.138, 0.994]), decimal=3)


def test_percus_yevick_verlet_weiss_initialization(mock_mixture):
    wavevector = [0.1, 0.2, 0.3]
    volume_fraction_total = 0.3
    pyvw = VerletWeis(wavevector, mock_mixture, volume_fraction_total)
    assert np.allclose(pyvw.wavevector, [0.1, 0.2, 0.3])
    assert pyvw.mixture == mock_mixture
    assert pyvw.volume_fraction_total == pytest.approx(volume_fraction_total * (1 - volume_fraction_total / 16))
    assert_array_almost_equal(pyvw.mixture.radius, np.array([0.994, 1.987]), decimal=3)


if __name__ == "__main__":
    pytest.main()



# type: ignore
import pytest
import numpy as np

from mixscatter.mixture import (
    Mixture,
    FlorySchulzMixture,
    GaussianMixture,
    UniformMixture,
    SingleComponent,
)


def test_mixture_initialization():
    radius = [100, 250]
    number_fraction = [0.5, 0.5]
    mixture = Mixture(radius, number_fraction)

    np.testing.assert_allclose(mixture.radius, np.array(radius))
    np.testing.assert_allclose(mixture.number_fraction, np.array(number_fraction))
    assert mixture.number_of_components == len(radius)


def test_mixture_normalization():
    radius = [100, 250]
    number_fraction = [1.0, 2.0]
    mixture = Mixture(radius, number_fraction, normalize_number_fraction=True)

    expected_number_fraction = np.array([1.0, 2.0]) / 3.0
    np.testing.assert_allclose(mixture.number_fraction, expected_number_fraction)


def test_mixture_no_normalization():
    radius = [100, 250]
    number_fraction = [1.0, 2.0]
    mixture = Mixture(radius, number_fraction, normalize_number_fraction=False)

    # Test that number_fraction remains unchanged
    np.testing.assert_array_equal(mixture.number_fraction, np.array(number_fraction))


def test_mixture_moment():
    radius = [100, 250]
    number_fraction = [0.5, 0.5]
    mixture = Mixture(radius, number_fraction)

    moment_order_1 = mixture.moment(1)
    expected_moment_order_1 = np.sum(np.array(number_fraction) * np.array(radius) ** 1)
    assert moment_order_1 == expected_moment_order_1


def test_mixture_central_moment():
    radius = [100, 250]
    number_fraction = [0.5, 0.5]
    mixture = Mixture(radius, number_fraction)

    central_moment_order_2 = mixture.central_moment(2)
    expected_central_moment_order_2 = np.sum(number_fraction * (np.array(radius) - mixture.mean) ** 2)
    assert central_moment_order_2 == expected_central_moment_order_2


def test_mixture_mean():
    radius = [100, 250]
    number_fraction = [0.5, 0.5]
    mixture = Mixture(radius, number_fraction)

    expected_mean = np.sum(np.array(number_fraction) * np.array(radius))
    assert mixture.mean == expected_mean


def test_mixture_variance():
    radius = [100, 250]
    number_fraction = [0.5, 0.5]
    mixture = Mixture(radius, number_fraction)

    expected_variance = np.sum(number_fraction * (np.array(radius) - mixture.mean) ** 2)
    assert mixture.variance == expected_variance


def test_mixture_polydispersity():
    radius = [100, 250]
    number_fraction = [0.5, 0.5]
    mixture = Mixture(radius, number_fraction)

    expected_polydispersity = np.sqrt(mixture.variance) / mixture.mean
    assert mixture.polydispersity == expected_polydispersity


def test_mixture_setters():
    radius = [100, 250]
    number_fraction = [0.5, 0.5]
    mixture = Mixture(radius, number_fraction)

    np.testing.assert_array_equal(mixture.radius, np.array(radius))
    np.testing.assert_array_equal(mixture.number_fraction, np.array(number_fraction))

    new_radius = [50, 500]
    new_number_fraction = [0.2, 0.8]
    mixture.radius = new_radius
    mixture.number_fraction = new_number_fraction
    np.testing.assert_array_equal(mixture.radius, np.array(new_radius))
    np.testing.assert_array_equal(mixture.number_fraction, np.array(new_number_fraction))


def test_mixture_setters_invalid_shape():
    radius = [100, 250]
    number_fraction = [0.5, 0.5]
    mixture = Mixture(radius, number_fraction)

    new_radius = [50, 500, 700]
    new_number_fraction = [0.2, 0.6, 0.2]

    with pytest.raises(ValueError, match="The new radius array must be of same shape as the current radius array."):
        mixture.radius = new_radius

    with pytest.raises(ValueError, match="The new radius array must be of same shape as the current radius array."):
        mixture.number_fraction = new_number_fraction


def test_single_component_initialization():
    radius = 500
    mixture = SingleComponent(radius)
    assert np.allclose(mixture.radius, [radius])
    assert np.allclose(mixture.number_fraction, [1.0])


def test_flory_schulz_mixture_initialization():
    number_of_components = 16
    mean_radius = 100
    shape_parameter = 99
    mixture = FlorySchulzMixture(number_of_components, mean_radius, shape_parameter)
    assert mixture.number_of_components == number_of_components
    assert np.isclose(mixture.mean, mean_radius)
    assert np.isclose(mixture.polydispersity, 1 / np.sqrt(shape_parameter + 1))


def test_gaussian_mixture_initialization():
    number_of_components = 16
    mean_radius = 100
    standard_deviation = 10
    mixture = GaussianMixture(number_of_components, mean_radius, standard_deviation)
    assert mixture.number_of_components == number_of_components
    assert np.isclose(mixture.mean, mean_radius)
    assert np.isclose(mixture.polydispersity, standard_deviation / mean_radius)


def test_gaussian_mixture_truncate():
    number_of_components = 16
    mean_radius = 100
    standard_deviation = 40
    mixture_truncate_negatives = GaussianMixture(
        number_of_components, mean_radius, standard_deviation, truncate="negatives"
    )
    mixture_truncate_symmetric = GaussianMixture(
        number_of_components, mean_radius, standard_deviation, truncate="symmetric"
    )
    mixture_truncate_no = GaussianMixture(number_of_components, mean_radius, standard_deviation, truncate="no")
    assert mixture_truncate_negatives.number_of_components < number_of_components
    assert np.all(mixture_truncate_negatives.radius >= 0)
    assert np.all(
        mixture_truncate_symmetric.radius - mixture_truncate_symmetric.mean <= mixture_truncate_symmetric.mean
    )
    assert mixture_truncate_symmetric.number_of_components < number_of_components
    assert 2 * (number_of_components - mixture_truncate_negatives.number_of_components) == (
        number_of_components - mixture_truncate_symmetric.number_of_components
    )

    assert mixture_truncate_no.number_of_components == number_of_components


def test_gaussian_mixture_invalid_truncate():
    number_of_components = 16
    mean_radius = 100
    standard_deviation = 10

    with pytest.raises(ValueError, match="Parameter 'truncate' must be 'no', 'negatives', or 'symmetric'."):
        GaussianMixture(number_of_components, mean_radius, standard_deviation, truncate="invalid")


def test_uniform_mixture_initialization():
    number_of_components = 16
    lower_bound = 50
    upper_bound = 150
    mixture = UniformMixture(number_of_components, lower_bound, upper_bound)
    assert mixture.number_of_components == number_of_components
    assert np.isclose(mixture.mean, (upper_bound + lower_bound) / 2)
    assert np.isclose(mixture.variance, (upper_bound - lower_bound) ** 2 / 12)


if __name__ == "__main__":
    pytest.main()



"""
This module provides functionalities to represent and create mixtures of spherical particles
with various size distributions. It includes classes and functions to handle different
distributions like single-component, Flory-Schulz, Gaussian, and uniform distributions.

Classes:
    Mixture: Base class representing an N-component mixture of spherical particles.
    FlorySchulzMixture: Represents a Schulz-Flory distributed mixture.
    GaussianMixture: Represents a Gaussian distributed mixture.
    UniformMixture: Represents a uniformly distributed mixture.
    SingleComponent: Represents a pseudo one-component system.
"""

import logging

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.special import roots_genlaguerre, eval_genlaguerre, roots_hermite, roots_legendre

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)


class Mixture:
    """Representation of the size distribution of an N-component mixture of spherical particles.

    Such a mixture can be interpreted as a collection of spheres with a discrete probability
    distribution of the radii, where the number fraction of each species describes their
    respective weight.

    Args:
        radius: The physical, geometric radius of each component.
        number_fraction: Number fraction of each component.
        normalize_number_fraction: If True, `number_fraction` will be explicitly normalized
                                   such that the sum over all elements equals 1.
    """

    def __init__(self, radius: ArrayLike, number_fraction: ArrayLike, normalize_number_fraction: bool = True) -> None:
        self._radius: NDArray[np.float64] = np.asarray(radius, dtype=np.float64)
        self._number_fraction: NDArray[np.float64] = np.asarray(number_fraction, dtype=np.float64)

        if normalize_number_fraction:
            self._number_fraction /= np.sum(self._number_fraction)

        self._number_of_components: int = len(self._radius)

    @property
    def radius(self) -> NDArray[np.float64]:
        """
        Get an array of the radii of each component.

        Returns:
            An array of the radii.
        """
        return self._radius

    @radius.setter
    def radius(self, radius_array: ArrayLike) -> None:
        """
        Set new `radius`.

        Args:
            radius_array: New radius array of the same shape as the current radius array.
        """
        new_array: NDArray[np.float64] = np.asarray(radius_array, dtype=np.float64)
        if self._radius.shape != new_array.shape:
            raise ValueError("The new radius array must be of same shape as the current radius array.")
        self._radius = new_array

    @property
    def number_fraction(self) -> NDArray[np.float64]:
        """
        Get an array of the number fractions of each component

        Returns:
            An array of the number fractions.
        """
        return self._number_fraction

    @number_fraction.setter
    def number_fraction(self, number_fraction_array: ArrayLike) -> None:
        """
        Set new `number_fraction`.

        Args:
            number_fraction_array:
        """
        new_array: NDArray[np.float64] = np.asarray(number_fraction_array, dtype=np.float64)
        if self._number_fraction.shape != new_array.shape:
            raise ValueError("The new radius array must be of same shape as the current radius array.")
        self._number_fraction = new_array

    @property
    def number_of_components(self) -> int:
        """
        Get the number of components of the mixture.

        Returns:
            The number of components.
        """
        return self._number_of_components

    def moment(self, order: int) -> float:
        """Calculate the moment of the specified order.

        Args:
            order: Order of the moment to calculate.

        Returns:
            The moment of the specified order.
        """
        return float(np.sum(self.number_fraction * self.radius**order))

    def central_moment(self, order: int) -> float:
        """Calculate the central moment of the specified order.

        Args:
            order: Order of the central moment to calculate.

        Returns:
            The central moment of the specified order.
        """
        return float(np.sum(self.number_fraction * (self.radius - self.mean) ** order))

    @property
    def mean(self) -> float:
        """
        Calculate the first moment (mean) of the distribution.

        Returns:
            The mean of the distribution.
        """
        return self.moment(1)

    @property
    def variance(self) -> float:
        """
        Calculate the second central moment (variance) of the distribution.

        Returns:
            The variance of the distribution.
        """
        return self.central_moment(2)

    @property
    def polydispersity(self) -> float:
        """
        Calculate the polydispersity, which is the ratio of the standard deviation
        of the distribution to its mean.

        Returns:
            The polydispersity of the distribution.
        """
        return float(np.sqrt(self.variance)) / self.mean


class SingleComponent(Mixture):
    """Representation of a pseudo one-component system.

    Args:
        radius: Radius.
    """

    def __init__(self, radius: float) -> None:
        super().__init__([radius], [1.0])


class FlorySchulzMixture(Mixture):
    """Representation of a Schulz-Flory distributed mixture.

    Args:
        number_of_components: Number of components in the mixture.
        mean_radius: Mean radius of the distribution.
        shape_parameter: Flory parameter controlling the distribution shape.
    """

    def __init__(self, number_of_components: int, mean_radius: float, shape_parameter: float) -> None:
        roots, weights = self._flory_schulz_weights(
            int(number_of_components), float(mean_radius), float(shape_parameter)
        )
        super().__init__(roots, weights)

    @classmethod
    def _flory_schulz_weights(
        cls, number_of_components: int, mean_radius: float, shape_parameter: float
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        roots, weights = cls._generalized_laguerre_weights(number_of_components, shape_parameter)
        scaled_roots = roots * mean_radius / (shape_parameter + 1.0)
        weights /= np.sum(weights)
        return scaled_roots, weights

    @staticmethod
    def _generalized_laguerre_weights(order: int, exponent: float) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        roots, *_ = roots_genlaguerre(order, exponent)
        values_at_roots = eval_genlaguerre(order + 1, exponent, roots)
        abs_values_at_roots = np.abs(values_at_roots)
        log_weights = np.log(roots) - 2.0 * np.log(abs_values_at_roots)
        weights = np.exp(log_weights)
        return roots, weights


class GaussianMixture(Mixture):
    """Representation of a Gaussian distributed mixture.

    Args:
        number_of_components: Number of components in the mixture.
        mean_radius: Mean radius of the distribution.
        standard_deviation: Standard deviation of the distribution.
        truncate:
            Method to handle unphysical radii. Options are 'no', 'negatives', and 'symmetric'.

            - 'no': No truncation

            - 'negatives': Truncate negative radii.

            - 'symmetric': Truncate negative radii and force the distribution to stay symmetric.
    """

    def __init__(
        self, number_of_components: int, mean_radius: float, standard_deviation: float, truncate: str = "negatives"
    ) -> None:
        if truncate not in ["no", "negatives", "symmetric"]:
            raise ValueError("Parameter 'truncate' must be 'no', 'negatives', or 'symmetric'.")

        roots, weights = self._gaussian_weights(
            int(number_of_components), float(mean_radius), float(standard_deviation)
        )

        if truncate == "negatives":
            roots_to_remove = roots < 0
        elif truncate == "symmetric":
            roots_to_remove = np.abs(roots - mean_radius) > mean_radius
        else:
            roots_to_remove = np.full_like(roots, False, dtype=bool)

        if np.any(roots_to_remove):
            roots_to_keep = ~roots_to_remove
            number_of_components_new = np.count_nonzero(roots_to_keep)
            roots = roots[roots_to_keep]
            weights = weights[roots_to_keep]

            logger.warning(
                f"Components with unphysical radii were removed. Number of components "
                f"was reduced from {number_of_components} to {number_of_components_new}."
            )

        super().__init__(roots, weights)

    @staticmethod
    def _gaussian_weights(
        number_of_components: int, mean_radius: float, standard_deviation: float
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        roots, weights = roots_hermite(number_of_components)
        scaled_roots = roots * np.sqrt(2) * standard_deviation + mean_radius
        weights /= np.sum(weights)
        return scaled_roots, weights


class UniformMixture(Mixture):
    """Representation of a uniformly distributed mixture.

    Args:
        number_of_components: Number of components in the mixture.
        lower_bound: Lower bound of the distribution domain.
        upper_bound: Upper bound of the distribution domain.
    """

    def __init__(self, number_of_components: int, lower_bound: float, upper_bound: float) -> None:
        roots, weights = self._uniform_weights(int(number_of_components), float(upper_bound), float(lower_bound))
        super().__init__(roots, weights)

    @staticmethod
    def _uniform_weights(
        number_of_components: int, lower_bound: float, upper_bound: float
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        roots, weights = roots_legendre(number_of_components)
        scaled_roots = upper_bound * 0.5 * (roots + 1) - lower_bound * 0.5 * (roots - 1)
        weights /= np.sum(weights)
        return scaled_roots, weights



"""Core module of mixscatter."""

import numpy as np

from typing import Any, Protocol

from numpy.typing import NDArray

__all__ = ["measurable_structure_factor", "measurable_intensity", "measurable_diffusion_coefficient"]


# noinspection PyPropertyDefinition


class MixtureLike(Protocol):  # pragma: no cover
    @property
    def radius(self) -> NDArray[np.float64]: ...

    @property
    def number_fraction(self) -> NDArray[np.float64]: ...


# noinspection PyPropertyDefinition
class LiquidStructureLike(Protocol):  # pragma: no cover
    """A protocol which defines an interface for a class which behaves like LiquidStructure"""

    @property
    def number_weighted_partial_structure_factor(self) -> NDArray[np.float64]: ...


# noinspection PyPropertyDefinition
class ScatteringModelLike(Protocol):  # pragma: no cover
    """A protocol which defines an interface for a class which behaves like ScatteringModel"""

    mixture: MixtureLike

    @property
    def amplitude(self) -> NDArray[np.float64]: ...

    @property
    def average_square_amplitude(self) -> Any: ...

    @property
    def average_square_forward_amplitude(self) -> Any: ...

    @property
    def average_form_factor(self) -> Any: ...


def measurable_intensity(
    liquid_structure: LiquidStructureLike,
    scattering_model: ScatteringModelLike,
    scale: float = 1.0,
    background: float = 0.0,
) -> Any:
    # noinspection PyShadowingNames
    """
    The measurable scattered intensity.

    Args:
        liquid_structure:
            Object instance with an interface similar to a LiquidStructure object.
        scattering_model:
            Object instance with an interface similar to a ScatteringModel object.
        scale:
            Scale the intensity by a factor.
        background:
            Add a constant background.

    Returns:
        Measurable scattered intensity.

    Examples:
        >>> import mixscatter as ms
        >>> import numpy as np

        >>> wavevector = np.linspace(0.005, 0.05, 100)
        >>> mixture = ms.mixture.Mixture([100, 200], [0.2, 0.8])

        >>> liquid_structure = ms.liquidstructure.PercusYevick(wavevector, mixture, volume_fraction_total=0.3)
        >>> scattering_model = ms.scatteringmodel.SimpleSphere(wavevector, mixture, contrast=1.0)

        >>> intensity = measurable_intensity(liquid_structure, scattering_model, scale=1e5,
        ...                                  background=1e3)
    """
    average_intensity = np.einsum(
        "iq, jq, ijq->q",
        scattering_model.amplitude,
        scattering_model.amplitude,
        liquid_structure.number_weighted_partial_structure_factor,
    )
    normalized_intensity = average_intensity / scattering_model.average_square_forward_amplitude
    return scale * normalized_intensity + background


def measurable_structure_factor(liquid_structure: LiquidStructureLike, scattering_model: ScatteringModelLike) -> Any:
    # noinspection PyShadowingNames
    """The measurable structure factor.

    This is the quotient of the measurable intensity of a system with interactions and the
    equivalent intensity of a system without interactions.

    Args:
        liquid_structure:
            Object instance with an interface similar to a LiquidStructure object.
        scattering_model:
            Object instance with an interface similar to a ScatteringModel object.

    Returns:
        Measurable structure factor.

    Examples:
        >>> import mixscatter as ms
        >>> import numpy as np

        >>> wavevector = np.linspace(0.005, 0.05, 100)
        >>> mixture = ms.mixture.Mixture([100, 200], [0.2, 0.8])

        >>> liquid_structure = ms.liquidstructure.PercusYevick(wavevector, mixture, volume_fraction_total=0.3)
        >>> scattering_model = ms.scatteringmodel.SimpleSphere(wavevector, mixture, contrast=1.0)

        >>> structure_factor = measurable_structure_factor(liquid_structure, scattering_model)
    """
    return measurable_intensity(liquid_structure, scattering_model) / scattering_model.average_form_factor


def measurable_diffusion_coefficient(
    scattering_model: ScatteringModelLike, thermal_energy: float, viscosity: float
) -> Any:
    # noinspection PyShadowingNames
    """
    The measurable Stokes-Einstein diffusion coefficient of a dilute suspension.

    Args:
        scattering_model:
            Object instance with an interface similar to a ScatteringModel object.
        thermal_energy:
            Thermal energy.
        viscosity:
            Viscosity of the surrounding medium.

    Returns:
        Measurable diffusion coefficient.

    Examples:
        >>> import mixscatter as ms
        >>> import numpy as np

        >>> wavevector = np.linspace(0.005, 0.05, 100)
        >>> mixture = ms.mixture.Mixture([100, 200], [0.2, 0.8])

        >>> scattering_model = ms.scatteringmodel.SimpleSphere(wavevector, mixture, contrast=1.0)
        >>> D_0 = measurable_diffusion_coefficient(
        ...         scattering_model, thermal_energy=1.0, viscosity=1.0)
    """
    weighted_inverse_radius = np.sum(
        scattering_model.mixture.number_fraction[:, np.newaxis]
        * scattering_model.amplitude**2
        / scattering_model.mixture.radius[:, np.newaxis],
        axis=0,
    )
    weighted_inverse_radius /= scattering_model.average_square_amplitude
    prefactor = thermal_energy / (6.0 * np.pi * viscosity)
    return prefactor * weighted_inverse_radius



# type: ignore
import pytest
import numpy as np

from mixscatter.core import measurable_intensity, measurable_structure_factor, measurable_diffusion_coefficient


@pytest.fixture
def mock_structure_factor(mocker):
    sf = mocker.Mock()
    sf.number_weighted_partial_structure_factor = np.random.rand(2, 2, 100)
    return sf


@pytest.fixture
def mock_form_factor(mocker):
    ff = mocker.Mock()
    ff.mixture = mocker.Mock()
    ff.mixture.number_fraction = np.random.rand(2)
    ff.mixture.radius = np.random.rand(2)
    ff.amplitude = np.random.rand(2, 100)
    ff.average_square_amplitude = np.random.rand(100)
    ff.average_square_forward_amplitude = np.random.rand(100)
    ff.average_form_factor = np.random.rand(100)
    return ff


def test_measurable_intensity_default(mock_structure_factor, mock_form_factor):
    scale = 1.0
    background = 0.0

    result = measurable_intensity(mock_structure_factor, mock_form_factor)

    expected_intensity = np.einsum(
        "iq, jq, ijq->q",
        mock_form_factor.amplitude,
        mock_form_factor.amplitude,
        mock_structure_factor.number_weighted_partial_structure_factor,
    )
    expected_normalized_intensity = expected_intensity / mock_form_factor.average_square_forward_amplitude
    expected_result = scale * expected_normalized_intensity + background

    np.testing.assert_allclose(result, expected_result)


def test_measurable_intensity_scaled(mock_structure_factor, mock_form_factor):
    scale = 1e5
    background = 1e3

    result = measurable_intensity(mock_structure_factor, mock_form_factor, scale=scale, background=background)

    expected_intensity = np.einsum(
        "iq, jq, ijq->q",
        mock_form_factor.amplitude,
        mock_form_factor.amplitude,
        mock_structure_factor.number_weighted_partial_structure_factor,
    )
    expected_normalized_intensity = expected_intensity / mock_form_factor.average_square_forward_amplitude
    expected_result = scale * expected_normalized_intensity + background

    np.testing.assert_allclose(result, expected_result)


def test_measurable_intensity_with_background(mock_structure_factor, mock_form_factor):
    scale = 1.0
    background = 100.0

    result = measurable_intensity(mock_structure_factor, mock_form_factor, background=background)

    expected_intensity = np.einsum(
        "iq, jq, ijq->q",
        mock_form_factor.amplitude,
        mock_form_factor.amplitude,
        mock_structure_factor.number_weighted_partial_structure_factor,
    )
    expected_normalized_intensity = expected_intensity / mock_form_factor.average_square_forward_amplitude
    expected_result = scale * expected_normalized_intensity + background

    np.testing.assert_allclose(result, expected_result)


def test_measurable_structure_factor(mock_structure_factor, mock_form_factor):
    # Calculate expected measurable intensity
    expected_intensity = np.einsum(
        "iq, jq, ijq->q",
        mock_form_factor.amplitude,
        mock_form_factor.amplitude,
        mock_structure_factor.number_weighted_partial_structure_factor,
    )
    expected_normalized_intensity = expected_intensity / mock_form_factor.average_square_forward_amplitude
    expected_measurable_intensity = expected_normalized_intensity  # scale is 1.0 and background is 0.0 by default

    # Calculate expected measurable structure factor
    expected_measurable_structure_factor = expected_measurable_intensity / mock_form_factor.average_form_factor

    # Call the function under test
    result = measurable_structure_factor(mock_structure_factor, mock_form_factor)

    # Assert that the result is as expected
    np.testing.assert_allclose(result, expected_measurable_structure_factor)


def test_measurable_diffusion_coefficient(mock_form_factor):
    thermal_energy = 1.0
    viscosity = 1.0

    # Calculate expected weighted inverse radius
    weighted_inverse_radius = np.sum(
        mock_form_factor.mixture.number_fraction[:, np.newaxis]
        * mock_form_factor.amplitude**2
        / mock_form_factor.mixture.radius[:, np.newaxis],
        axis=0,
    )
    weighted_inverse_radius /= mock_form_factor.average_square_amplitude

    # Calculate expected prefactor
    prefactor = thermal_energy / (6.0 * np.pi * viscosity)

    # Calculate expected diffusion coefficient
    expected_diffusion_coefficient = prefactor * weighted_inverse_radius

    # Call the function under test
    result = measurable_diffusion_coefficient(mock_form_factor, thermal_energy, viscosity)

    # Assert that the result is as expected
    np.testing.assert_allclose(result, expected_diffusion_coefficient)


if __name__ == "__main__":
    pytest.main()



"""Calculate form factors of multicomponent systems.

This module provides classes and functions for calculating form factors of multicomponent systems.
It includes definitions for various layer profiles, particles composed of these layers, and models
to calculate scattering properties for these particles.

Classes:
    LayerProfile: Defines the interface for layer profiles.
    EmptyProfile: Represents an empty layer profile.
    ConstantProfile: Represents a layer profile with constant contrast.
    LinearProfile: Represents a layer profile with linearly varying contrast.
    Particle: Represents a particle composed of multiple layers.
    ParticleBuilder: A builder class for constructing Particle instances.
    ScatteringModel: Calculates scattering properties for a list of particles.
    SimpleSphere: A convenience class for creating a scattering model of homogeneously scattering spheres.
    SimpleCoreShell: A convenience class for creating a scattering model of core-shell particles.
    SimpleGradient: A convenience class for creating a scattering model with gradient profiles.

Examples:
    Create a simple scattering model for spheres:

    >>> from mixscatter.mixture import Mixture
    >>> import numpy as np
    >>> wavevector = np.linspace(0.01, 1.0, 100)
    >>> mixture = Mixture(number_fraction=[0.5, 0.5], radius=[1.0, 2.0])
    >>> model = SimpleSphere(wavevector, mixture, contrast=1.0)
    >>> form_factor = model.average_form_factor

    Create a core-shell particle model:

    >>> model = SimpleCoreShell(wavevector, mixture, core_to_total_ratio=0.5, core_contrast=1.0, shell_contrast=0.5)
    >>> form_factor = model.average_form_factor

    Create a gradient profile particle model:

    >>> model = SimpleGradient(wavevector, mixture, center_contrast=1.0, boundary_contrast=0.5)
    >>> form_factor = model.average_form_factor
"""

from .scatteringmodel import (
    LayerProfile,
    EmptyProfile,
    ConstantProfile,
    LinearProfile,
    Particle,
    ParticleBuilder,
    ScatteringModel,
    SimpleSphere,
    SimpleCoreShell,
    SimpleGradient,
)

__all__ = [
    "LayerProfile",
    "EmptyProfile",
    "ConstantProfile",
    "LinearProfile",
    "Particle",
    "ParticleBuilder",
    "ScatteringModel",
    "SimpleSphere",
    "SimpleCoreShell",
    "SimpleGradient",
]



"""
Module `liquidstructure` provides classes for representing the liquid structure properties of
interacting multicomponent systems of spherical particles.

Classes:
    LiquidStructure: Abstract base class representing liquid structure properties.
    PercusYevick: Class for hard-sphere potential in the Percus-Yevick approximation.
    VerletWeis: Class for hard-sphere potential in the Percus-Yevick approximation
        with Verlet-Weiss correction.

References:
    - P. Salgi, R. Rajagopalan, "Polydispersity in colloids: implications to static structure and scattering",
      Adv. Colloid Interface Sci. 43, 169-288 (1993), https://doi.org/10.1016/0001-8686(93)80017-6
    - A. Vrij, "Mixtures of hard spheres in the Percus–Yevick approximation. Light scattering at finite angles",
      J. Chem. Phys. 71, 3267-3270 (1979), https://doi.org/10.1063/1.438756
    - J. Diaz Maier, K. Gaus, J. Wagner, "Measurable structure factors of dense dispersions containing polydisperse,
      optically inhomogeneous particles", arXiv:2404.03470 [cond-mat.soft], https://doi.org/10.48550/arXiv.2404.03470
"""

from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any, Protocol

import numpy as np
from numpy.typing import ArrayLike, NDArray


# noinspection PyPropertyDefinition
class MixtureLike(Protocol):  # pragma: no cover
    """
    A protocol defining the interface for mixture-like objects.

    Properties:
        number_fraction (NDArray[np.float64]): Number fractions of the components in the mixture.
        radius (NDArray[np.float64]): Radii of the components in the mixture.
        number_of_components (int): The number of different components in the mixture.

    Methods:
        moment(order: int) -> float: Calculate the specified moment of the size distribution.
    """

    @property
    def number_fraction(self) -> NDArray[np.float64]: ...

    @property
    def radius(self) -> NDArray[np.float64]: ...

    @radius.setter
    def radius(self, radius_array: ArrayLike) -> Any: ...

    @property
    def number_of_components(self) -> int: ...

    def moment(self, order: int) -> float: ...


class LiquidStructure(ABC):
    """
    Representation of the liquid structure properties of an interacting multicomponent system of
    spherical particles.

    Specialized classes inherit from this base class and only need to implement the
    'partial_direct_correlation_function' attribute.

    Attributes:
        wavevector (NDArray[np.float64]): Scattering wavevector.
        mixture (MixtureLike): Mixture object.

    References:
        - P. Salgi, R. Rajagopalan, "Polydispersity in colloids: implications to static structure and scattering",
          Adv. Colloid Interface Sci. 43, 169-288 (1993), https://doi.org/10.1016/0001-8686(93)80017-6
        - A. Vrij, "Mixtures of hard spheres in the Percus–Yevick approximation. Light scattering at finite angles",
          J. Chem. Phys. 71, 3267-3270 (1979), https://doi.org/10.1063/1.438756
        - J. Diaz Maier, K. Gaus, J. Wagner, "Measurable structure factors of dense dispersions containing polydisperse,
          optically inhomogeneous particles", arXiv:2404.03470 [cond-mat.soft],
          https://doi.org/10.48550/arXiv.2404.03470
    """

    def __init__(self, wavevector: ArrayLike, mixture: MixtureLike):
        """
        Initialize an instance based on the specified wavevectors and for a given mixture.

        Args:
            wavevector (ArrayLike): Scattering wavevector.
            mixture (MixtureLike): Mixture object.
        """
        self.wavevector: NDArray[np.float64] = np.asarray(wavevector, dtype=np.float64)
        self.mixture: MixtureLike = mixture

    @abstractmethod
    @cached_property
    def partial_direct_correlation_function(self) -> NDArray[np.float64]:  # pragma: no cover
        """
        Matrix of the partial direct correlation functions weighted by the total number density.

        Returns:
            NDArray[np.float64]: Partial direct correlation function matrix.
        """
        ...

    @cached_property
    def number_weighted_partial_direct_correlation_function(self) -> Any:
        """
        The partial direct correlation function matrix `c_ij` weighted by the square root of the
        product of the number fractions `x_i` and `x_j`.

        Returns:
            NDArray[np.float64]: Matrix of the weighted direct partial correlation functions.
        """
        c_weighted_ijq = np.einsum(
            "i, ijq, j->ijq",
            np.sqrt(self.mixture.number_fraction),
            self.partial_direct_correlation_function,
            np.sqrt(self.mixture.number_fraction),
            optimize="greedy",
        )
        return c_weighted_ijq

    @cached_property
    def partial_structure_factor(self) -> NDArray[np.float64]:
        """
        The partial structure factor matrix.

        Returns:
            NDArray[np.float64]: Matrix of the partial structure factors.
        """
        c_weighted_ijq = self.number_weighted_partial_direct_correlation_function
        c_weighted_qij = np.moveaxis(c_weighted_ijq, -1, 0)
        unity_tensor_qij = np.eye(self.mixture.number_of_components)[np.newaxis, :, :]
        S_inv_qij = unity_tensor_qij - c_weighted_qij
        S_qij = np.linalg.solve(S_inv_qij, unity_tensor_qij)
        S_ijq = np.moveaxis(S_qij, 0, -1)
        return S_ijq

    @cached_property
    def number_weighted_partial_structure_factor(self) -> Any:
        """
        The partial direct correlation function matrix `S_ij` weighted by the square root of the
        product of the number fractions `x_i` and `x_j`.

        Returns:
            NDArray[np.float64]: Matrix of the weighted direct partial correlation functions.
        """
        S_weighted_ijq = np.einsum(
            "i, ijq, j->ijq",
            np.sqrt(self.mixture.number_fraction),
            self.partial_structure_factor,
            np.sqrt(self.mixture.number_fraction),
            optimize="greedy",
        )
        return S_weighted_ijq

    @cached_property
    def average_structure_factor(self) -> Any:
        """
        The sum of the number weighted partial structure factors over all species.

        The number-number structure factor for all species, regardless their individual size.

        Returns:
            NDArray[np.float64]: Average structure factor.
        """
        return np.einsum("ijq->q", self.number_weighted_partial_structure_factor)

    @cached_property
    def compressibility_structure_factor(self) -> Any:
        """
        The "compressibility" or "Kirkwood-Buff" structure factor.

        This structure factor gives the correct isothermal compressibility from S(Q->0) according
        to the Kirkwood-Buff-theory for multicomponent systems.

        Returns:
            NDArray[np.float64]: "Compressibility" structure factor.
        """
        S_weighted_ijq = self.number_weighted_partial_structure_factor
        S_weighted_qij = np.moveaxis(S_weighted_ijq, -1, 0)
        S_weighted_inverse_qij = np.linalg.inv(S_weighted_qij)
        return 1.0 / np.einsum(
            "i, j, qij->q",
            self.mixture.number_fraction,
            self.mixture.number_fraction,
            S_weighted_inverse_qij,
        )


class PercusYevick(LiquidStructure):
    """
    Hard-sphere potential in the Percus-Yevick approximation.

    References:
        - A. Vrij, "Mixtures of hard spheres in the Percus–Yevick approximation. Light scattering at finite angles",
          J. Chem. Phys. 71, 3267-3270 (1979), https://doi.org/10.1063/1.438756
        - T. Voigtmann, "Mode Coupling Theory of the Glass Transition in Binary Mixtures",
          PhD Thesis, TU München, https://mediatum.ub.tum.de/603008
        - J. Diaz Maier, K. Gaus, J. Wagner, "Measurable structure factors of dense dispersions containing polydisperse,
          optically inhomogeneous particles", arXiv:2404.03470 [cond-mat.soft],
          https://doi.org/10.48550/arXiv.2404.03470
    """

    def __init__(self, wavevector: ArrayLike, mixture: MixtureLike, volume_fraction_total: float):
        """
        Initialize an instance based on the specified wavevectors and for a given mixture.

        Args:
            wavevector (ArrayLike): Scattering wavevector.
            mixture (MixtureLike): Mixture object.
            volume_fraction_total (float): Total volume fraction.
        """
        self.volume_fraction_total: float = volume_fraction_total
        super().__init__(wavevector, mixture)

    @cached_property
    def partial_direct_correlation_function(self) -> Any:
        """
        Calculate the partial direct correlation function matrix `c_ij(q)` for the system using the
        Percus-Yevick approximation.

        Returns:
            NDArray[np.float64]: Partial direct correlation function matrix `c_ij(q)` times the total number density.
        """
        average_volume = 4.0 / 3.0 * np.pi * self.mixture.moment(3)
        number_density_total = self.volume_fraction_total / average_volume
        number_density = number_density_total * self.mixture.number_fraction

        diameter = 2.0 * self.mixture.radius
        xi = [np.pi / 6.0 * np.sum(number_density * diameter**x) for x in range(1, 4)]
        diameter_mean = 0.5 * (diameter[:, np.newaxis] + diameter)
        diameter_product = diameter[:, np.newaxis] * diameter
        beta_hat_0 = (9.0 * xi[1] ** 2 + 3.0 * xi[0] * (1 - xi[2])) / (1 - xi[2]) ** 3
        a_i = (1.0 - xi[2] + 3.0 * diameter * xi[1]) / (1 - xi[2]) ** 2
        a_tilde_2 = np.sum(number_density * a_i**2)
        A_ij = (diameter_mean * (1 - xi[2]) + 1.5 * diameter_product * xi[1]) / (1 - xi[2]) ** 2
        B_ij = 1.0 / (1 - 0 - xi[2]) - beta_hat_0 * diameter_product
        D_ij = (6.0 * xi[1] + 12.0 * diameter_mean * (xi[0] + 3.0 * xi[1] ** 2 / (1 - xi[2]))) / (1 - xi[2]) ** 2
        wavevector_reduced = self.mixture.radius[:, np.newaxis] * self.wavevector
        wavevector_reciprocal = 1.0 / self.wavevector
        sin_iq = np.sin(wavevector_reduced)
        cos_iq = np.cos(wavevector_reduced)

        c_ijq = 4.0 * np.pi * a_tilde_2 * sin_iq[:, np.newaxis] * sin_iq
        c_ijq *= wavevector_reciprocal

        c_ijq -= (
            2.0
            * np.pi
            * a_tilde_2
            * (
                cos_iq[:, np.newaxis] * sin_iq * diameter[:, np.newaxis, np.newaxis]
                + sin_iq[:, np.newaxis] * cos_iq * diameter[np.newaxis, :, np.newaxis]
            )
        )
        c_ijq *= wavevector_reciprocal

        c_ijq += (
            D_ij[:, :, np.newaxis] * sin_iq[:, np.newaxis] * sin_iq
            + np.pi * a_tilde_2 * cos_iq[:, np.newaxis] * cos_iq * diameter_product[:, :, np.newaxis]
        )
        c_ijq *= wavevector_reciprocal

        c_ijq += B_ij[:, :, np.newaxis] * (cos_iq[:, np.newaxis] * sin_iq + sin_iq[:, np.newaxis] * cos_iq)
        c_ijq *= wavevector_reciprocal

        c_ijq += A_ij[:, :, np.newaxis] * (sin_iq[:, np.newaxis] * sin_iq - cos_iq[:, np.newaxis] * cos_iq)
        c_ijq *= -4.0 * np.pi * wavevector_reciprocal**2
        c_ijq *= number_density_total
        return c_ijq


class VerletWeis(PercusYevick):
    """
    Hard-sphere potential in the Percus-Yevick approximation employing the Verlet-Weiss correction.

    Uses the PercusYevick class with an effective volume fraction and an effective radius.

    References:
        - E. W. Grundke and D. Henderson, "Distribution functions of multi-component fluid mixtures of hard spheres",
          Mol. Phys. 24, 269-281 (1972), https://doi.org/10.1080/00268977200101431
    """

    def __init__(self, wavevector: ArrayLike, mixture: MixtureLike, volume_fraction_total: float) -> None:
        """
        Initialize an instance based on the specified wavevectors and for a given mixture.

        Args:
            wavevector (ArrayLike): Scattering wavevector.
            mixture (MixtureLike): Mixture object.
            volume_fraction_total (float): Total volume_fraction.
        """
        effective_volume_fraction = volume_fraction_total * (1.0 - volume_fraction_total / 16.0)
        effective_radius = mixture.radius * (effective_volume_fraction / volume_fraction_total) ** (1.0 / 3.0)
        mixture.radius = effective_radius
        super().__init__(wavevector, mixture, volume_fraction_total=effective_volume_fraction)



# noinspection PyShadowingNames
"""
Calculate scattering functions of multicomponent mixtures of interacting spherical scatterers \
within the Born approximation.

Examples:
    >>> import numpy as np
    >>> import mixscatter as ms

    >>> particle_mixture = ms.mixture.Mixture(radius=[100, 250], number_fraction=[0.5, 0.5])

    >>> wavevector = np.linspace(1e-3, 6e-2, 4096)
    >>> scattering_model = ms.scatteringmodel.SimpleSphere(wavevector, particle_mixture, contrast=1.0)
    >>> structure_factor = ms.liquidstructure.PercusYevick(
    ... wavevector, particle_mixture, volume_fraction_total=0.3
    ... )

    >>> measurable_intensity = ms.measurable_intensity(structure_factor, scattering_model)
    >>> measurable_structure_factor = ms.measurable_structure_factor(structure_factor, scattering_model)
    >>> average_form_factor = scattering_model.average_form_factor
"""

from importlib.metadata import version, PackageNotFoundError

from mixscatter.core import (
    measurable_intensity,
    measurable_structure_factor,
    measurable_diffusion_coefficient,
    LiquidStructureLike,
    ScatteringModelLike,
)
from mixscatter import liquidstructure, scatteringmodel, mixture

__all__ = [
    "measurable_intensity",
    "measurable_structure_factor",
    "measurable_diffusion_coefficient",
    "LiquidStructureLike",
    "ScatteringModelLike",
    "liquidstructure",
    "scatteringmodel",
    "mixture",
]

try:
    __version__ = version("mixscatter")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    pass
